"""Mutable set of streams."""

from collections.abc import MutableSet
from typing import Any, Callable, Iterator, Self

from ytpb.format_spec import query_items, QueryFunction
from ytpb.types import AudioOrVideoStream, AudioStream, SetOfStreams, VideoStream

__all__ = ("Streams",)


def stream_comparison_function(stream: AudioOrVideoStream):
    match stream:
        case AudioStream():
            return (stream.mime_type, stream.audio_sampling_rate)
        case VideoStream():
            return (stream.mime_type, stream.height, stream.frame_rate)


class Streams(MutableSet):
    """Represents a mutable set of
    :class:`ytpb.representations.RepresentationInfo` objects."""

    def __init__(self, iterable: list[AudioOrVideoStream] | None = None) -> None:
        self._elements = set()
        for value in iterable or []:
            if not isinstance(value, AudioOrVideoStream):
                raise ValueError
            self._elements.add(value)

    @classmethod
    def from_dicts(cls, dicts: list[dict]):
        streams = cls()
        stream: AudioOrVideoStream
        for stream_dict in dicts:
            if "audio" in stream_dict["mime_type"]:
                stream = AudioStream(**stream_dict)
            else:
                stream = VideoStream(**stream_dict)
            streams.add(stream)
        return streams

    def __len__(self):
        return len(self._elements)

    def __ior__(self, other: Self) -> Self:
        self.update(other)
        return self

    def __iter__(self) -> Iterator[AudioOrVideoStream]:
        return iter(self._elements)

    def __contains__(self, item: Any) -> bool:
        for stream in self:
            if stream.itag == item.itag:
                return True
        return False

    def add(self, value: AudioOrVideoStream):
        """Adds a stream."""
        self._elements.add(value)

    def discard(self, value: AudioOrVideoStream):
        """Removes a stream."""
        try:
            self._elements.remove(value)
        except KeyError:
            pass

    def get_by_itag(self, itag: str) -> AudioOrVideoStream | None:
        """Gets a stream by an itag value."""
        for stream in self:
            if stream.itag == itag:
                return stream
        return None

    def filter(self, predicate: Callable[[AudioOrVideoStream], bool]) -> SetOfStreams:
        """Filters streams by a predicate function.

        Args:
            predicate: A predicate function.

        Examples:
            Get only video streams::

              from ytpb.types import VideoStream
              playback.streams.filter(lambda x: x.type == "video")

        Returns:
            A new instance of this class with filtered streams.
        """
        return self.__class__(list(filter(predicate, self._elements)))

    def query(
        self,
        format_spec: str,
        functions: dict[str, QueryFunction] | None = None,
    ) -> list[AudioOrVideoStream]:
        """Queries streams by a format spec.

        Notes:
            Check for attributes is strict. This means that the following code
            will fail with :class:`.QueryError` because audio streams don't have
            the ``height`` attribute::

                audio_streams.query("height eq 1080")

        References:
            https://ytpb.readthedocs.io/en/latest/reference.html#format-spec

        Args:
            format_spec: A format spec.
            functions: Additional query functions to be used in a format
                spec. For built-in functions, see
                :const:`ytpb.format_spec.FUNCTIONS`.

        Returns:
            A list of queried streams.

        Raises:
            QueryError: If failed to query streams with the given format spec.
        """
        return query_items(format_spec, self._elements, functions)



import pytest

from ytpb.cli.utils.path import adjust_for_filename, AllowedCharacters


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "En direct : Titre de  la   vidéo — 24h/7 | Panorama, 360 / ? ",
            "En direct Titre de la vidéo — 24h-7 - Panorama, 360",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 - TBS NEWS DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_for_filename_as_is(title, expected):
    assert expected == adjust_for_filename(title, length=255)


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "En direct : Titre de  la   vidéo — 24h/7 | Panorama, 360 / ? ",
            "En-direct-Titre-de-la-vidéo—24h-7-Panorama,-360",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "【LIVE】新宿駅前の様子-Shinjuku,-Tokyo-JAPAN【ライブカメラ】-TBS-NEWS-DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_for_filename_with_separator(title, expected):
    assert expected == adjust_for_filename(title, separator="-", length=255)


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "En direct : Titre de  la   vidéo — 24h/7 | Panorama, 360 / ? ",
            "EndirectTitredelavideo--24h-7-Panorama,360",
        )
    ],
)
def test_adjust_for_filename_with_ascii_and_empty_separator(title, expected):
    assert expected == adjust_for_filename(
        title, separator="", length=255, characters=AllowedCharacters.ASCII
    )


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "En direct : Titre de  la   vidéo — 24h/7 | Panorama, 360 / ? ",
            "En direct Titre de la video -- 24h-7 - Panorama, 360",
        )
    ],
)
def test_adjust_for_filename_with_ascii_and_whitespace_separator(title, expected):
    assert expected == adjust_for_filename(
        title, separator=" ", length=255, characters=AllowedCharacters.ASCII
    )


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "En direct : Titre de  la   vidéo — 24h/7 | Panorama, 360 / ? ",
            "En-direct-Titre-de-la-video--24h-7-Panorama,-360",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "[(LIVE)]-Xin-Su-Yi-Qian-noYang-Zi--Shinjuku,-Tokyo-JAPAN[(raibukamera)]-TBS-NEWS-DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_for_filename_with_ascii_and_separator(title, expected):
    assert expected == adjust_for_filename(
        title, separator="-", length=255, characters=AllowedCharacters.ASCII
    )


@pytest.mark.parametrize(
    "title,expected,separator",
    [
        (
            "Jackman Maine - Newton Field Airport - Cam Left",
            "Jackman-Maine-Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine -  Newton Field Airport  -  Cam Left",
            "Jackman-Maine-Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine —  Newton Field Airport - Cam Left",
            "Jackman-Maine--Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine - Newton Field Airport - Cam Left",
            "Jackman_Maine-Newton_Field_Airport-Cam_Left",
            "_",
        ),
        (
            "Jackman Maine -  Newton Field Airport  -  Cam Left",
            "Jackman_Maine-Newton_Field_Airport-Cam_Left",
            "_",
        ),
        (
            "Jackman Maine — Newton Field Airport - Cam Left",
            "Jackman_Maine--Newton_Field_Airport-Cam_Left",
            "_",
        ),
    ],
)
def test_corner_cases_of_adjust_title_for_filename(title, expected, separator):
    assert expected == adjust_for_filename(
        title, separator=separator, characters=AllowedCharacters.ASCII
    )
    assert expected == adjust_for_filename(
        title, separator=separator, characters=AllowedCharacters.POSIX
    )


def test_adjust_for_filename_with_posix():
    title = "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG"

    expected = "LIVE_Xin_Su_Yi_Qian_noYang_Zi_Shinjuku_Tokyo_JAPAN-TBS_NEWS_DIG"
    assert expected == adjust_for_filename(
        title, separator="_", length=255, characters=AllowedCharacters.POSIX
    )

    expected = "LIVE-Xin-Su-Yi-Qian-noYang-Zi-Shinjuku-Tokyo-JAPAN-TBS-NEWS-DIG"
    assert expected == adjust_for_filename(
        title, separator="*", length=255, characters=AllowedCharacters.POSIX
    )


@pytest.mark.parametrize(
    "title,expected,length",
    [
        (
            "En direct : Titre de la vidéo",
            #  1              16
            # "En direct : Titre de la vidéo"
            "En_direct_Titre",
            16,
        ),
        (
            "En direct : Titre de la vidéo",
            #  1          12
            # "En direct : Titre de la vidéo"
            "En_direct",
            12,
        ),
    ],
)
def test_adjust_for_filename_with_length_and_separator(title, expected, length):
    assert expected == adjust_for_filename(
        title,
        separator="_",
        length=length,
        characters=AllowedCharacters.ASCII,
    )



from datetime import timezone
from pathlib import Path

import av
import pytest


def assert_number_of_streams(file_path: Path, expected: int):
    with av.open(file_path) as container:
        assert len(container.streams) == expected


def assert_approx_duration(file_path: Path, expected: float, abs: float = 2.2e-2):
    with av.open(file_path) as container:
        actual = container.duration / 1e6

    approx_expected = pytest.approx(expected, abs=abs)
    if actual != approx_expected:
        raise AssertionError(
            "Durations are not almost equal\n"
            f" - expected: {approx_expected}\n"
            f" + actual: {actual}"
        )


# See https://github.com/spulec/freezegun/issues/348.
def patched_freezgun_astimezone(self, tz=None):
    from freezegun.api import datetime_to_fakedatetime, real_datetime

    return datetime_to_fakedatetime(
        real_datetime.astimezone(self, timezone(self._tz_offset()))
    )



import click
import pytest

from ytpb.cli.formats import ALIASES, expand_aliases


@pytest.mark.parametrize(
    "aliases,spec,expected",
    [
        ({"a": "itag eq 140"}, "@a", "itag eq 140"),
        ({"a1": "itag eq 140"}, "@a1", "itag eq 140"),
        ({"a!": "itag eq 140"}, "@a!", "itag eq 140"),
        ({"a?": "itag eq 140"}, "@a?", "itag eq 140"),
        ({"a-a": "itag eq 140"}, "@a-a", "itag eq 140"),
        (
            {"a": "itag eq 140", "b": "format eq mp4"},
            "@a and @b",
            "itag eq 140 and format eq mp4",
        ),
        ({r"(\d+)\b": r"itag eq \1"}, "@140", "itag eq 140"),
        (
            {"a": "format eq mp4", "b": "frame_rate eq 60"},
            "[@a and @b] or [@a]",
            "[format eq mp4 and frame_rate eq 60] or [format eq mp4]",
        ),
    ],
)
def test_expand_non_nested_aliases(aliases: dict[str, str], spec: str, expected: str):
    assert expected == expand_aliases(spec, aliases)


@pytest.mark.parametrize(
    "aliases,spec,expected",
    [
        ({"a": "@aa", "aa": "itag eq 140"}, "@a", "itag eq 140"),
        (
            {"a": "@aa and @bb", "aa": "itag eq 140", "bb": "format eq mp4"},
            "@a",
            "itag eq 140 and format eq mp4",
        ),
        (
            {"full": "height eq 2160", "f": "@full"},
            "@f",
            "height eq 2160",
        ),
        (
            {"full": "height eq 2160", "f": "@full"},
            "[@full and frame_rate eq 60] or [@f]",
            "[height eq 2160 and frame_rate eq 60] or [height eq 2160]",
        ),
    ],
)
def test_expand_nested_aliases(aliases: dict[str, str], spec: str, expected: str):
    assert expected == expand_aliases(spec, aliases)


@pytest.mark.parametrize(
    "aliases,spec",
    [
        ({}, "@x"),
        ({}, "@x and @y"),
        ({"a": "@x"}, "@a"),
    ],
)
def test_expand_unknown_aliases(aliases: dict[str, str], spec: str):
    with pytest.raises(click.UsageError) as exc_info:
        expand_aliases(spec, aliases)
    assert "Unknown alias found: @x" == str(exc_info.value)


@pytest.mark.parametrize(
    "aliases,spec,error",
    [
        ({"a": "@a"}, "@a", "Cannot resolve circular alias @a in @a"),
        ({"a": "@b", "b": "@a"}, "@a", "Cannot resolve circular alias @a in @b"),
        (
            {"a": "@b", "b": "@c", "c": "@a and test"},
            "@a",
            "Cannot resolve circular alias @a in @c",
        ),
    ],
)
def test_expand_circular_dependent_aliases(
    aliases: dict[str, str], spec: str, error: str
):
    with pytest.raises(click.UsageError) as exc_info:
        expand_aliases(spec, aliases)
    assert error == str(exc_info.value)


@pytest.mark.parametrize(
    "spec,expected",
    [
        # Itag value aliases:
        ("@140", "itag = 140"),
        # Quality aliases:
        ("@720p", "height = 720"),
        ("@720p30", "[height = 720 and frame_rate = 30]"),
        # Quality with operator aliases:
        ("@<=720p", "height <= 720"),
        ("@>=720p", "height >= 720"),
        ("@=720p", "height = 720"),
        ("@==720p", "height == 720"),
        # Format aliases:
        ("@webm", "format = webm"),
        ("@mp4", "format = mp4"),
        # Codecs aliases:
        ("@mp4a", "codecs contains mp4a"),
        ("@avc1", "codecs contains avc1"),
        ("@vp9", "codecs = vp9"),
        # Named quality aliases:
        ("@low", "height = 144"),
        ("@medium", "height = 480"),
        ("@high", "height = 720"),
        ("@FHD", "height = 1080"),
        ("@2K", "height = 1440"),
        ("@4K", "height = 2160"),
    ],
)
def test_default_alias(spec, expected):
    assert expected == expand_aliases(spec, ALIASES)
    assert f"({expected})" == expand_aliases(f"({spec})", ALIASES)
    assert f"[{expected}]" == expand_aliases(f"[{spec}]", ALIASES)
